var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { Json } from "mvx-utils";
import { SignNonceStr } from "../Sign";
import { Token } from "../model/Token";
import { Md5 } from "ts-md5";
export var Code;
(function (Code) {
    Code[Code["Success"] = 200] = "Success";
    Code[Code["TokenExpire"] = 401] = "TokenExpire";
    Code[Code["ElseError"] = 500] = "ElseError";
})(Code || (Code = {}));
export class CodeError extends Error {
    constructor(err, code = Code.ElseError) {
        super();
        this.code = code;
        this.name = "CodeError";
        if (typeof err === 'string') {
            this.message = `${err}`;
        }
        else {
            this.message = `${err.message}`;
            this.stack = err.stack;
        }
    }
}
function newHttpOpt(...opts) {
    let opt = {
        header: new Map(),
        body: body => {
            throw new Error(`body encode not implement`);
        },
        method: "GET",
        token: tokenStr => { },
        httpCache: new class {
            get(key) {
                return ["", new CodeError("not found")];
            }
            has(key) {
                return false;
            }
            set(key, value) {
            }
        }
    };
    opts.forEach(f => {
        f(opt);
    });
    return opt;
}
export function WithHeader(header) {
    return (o) => {
        header.forEach((value, key) => {
            o.header.set(key, value);
        });
        return null;
    };
}
export function WithBearerToken() {
    return (o) => {
        o.token = (request, tokenStr) => {
            o.header.set("Authorization", `Bearer ${tokenStr}`);
        };
        return null;
    };
}
export function WithBodyToken(tokenKey = "token") {
    return (o) => {
        o.token = (body, tokenStr) => {
            body[tokenKey] = tokenStr;
        };
        return null;
    };
}
export function WithoutToken() {
    return (o) => {
        o.token = (body, tokenStr) => { };
        return null;
    };
}
export function WithReqId() {
    return (o) => {
        o.header.set("X-Req-Id", SignNonceStr());
        return null;
    };
}
export function WithBody(body) {
    return (o) => {
        o.body = body;
        return null;
    };
}
export function WithFormUrlEncodedBody() {
    return (o) => {
        o.header.set("X-Requested-With", "XMLHttpRequest");
        o.header.set('Content-Type', 'application/x-www-form-urlencoded');
        o.body = (body) => {
            const [json, err] = new Json().toJson(body);
            if (err) {
                return ["", err];
            }
            const args = JSON.parse(json);
            const ret = Object.keys(args).map((key) => {
                if (args[key] == null) {
                    return encodeURIComponent(key) + "=";
                }
                return encodeURIComponent(key) + "=" + encodeURIComponent(args[key]);
            }).join("&");
            return [ret, null];
        };
        return null;
    };
}
export function WithJsonBody() {
    return (o) => {
        o.header.set("X-Requested-With", "XMLHttpRequest");
        o.header.set('Content-Type', 'application/json;charset=utf8');
        o.body = (body) => {
            return new Json().toJson(body);
        };
        return null;
    };
}
export function WithHttpMethod(method) {
    return (o) => {
        o.method = method;
        return null;
    };
}
export function WithHttpCache(httpCache) {
    return (o) => {
        o.httpCache = httpCache;
        return null;
    };
}
export function SendNoToken(uri, request, resType, net, ...opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const opt = newHttpOpt(...opts);
        opt.token(request, yield net.getToken());
        const [body, err0] = opt.body(request);
        if (err0) {
            return [new resType(), new CodeError(err0.message)];
        }
        let retFunc = (res, err) => {
            if (err.code !== Code.Success) {
                return [new resType(), err];
            }
            if (res === "") {
                return [new resType(), new CodeError("response is empty")];
            }
            let [resT, err1] = new Json().fromJson(res, resType);
            if (err1) {
                return [new resType(), new CodeError(err1)];
            }
            return [resT, null];
        };
        const uniqueId = `${opt.method}@${net.getBaseUrl()}@${uri}@${Md5.hashStr(body)}`;
        if (opt.httpCache.has(uniqueId)) {
            return retFunc(...opt.httpCache.get(uniqueId));
        }
        const [res, err] = yield net.getHttpBuilder()
            .setUri(uri)
            .setMethod(opt.method)
            .setHeaders(opt.header)
            .setContent(body).build().send();
        opt.httpCache.set(uniqueId, [res, err]);
        return retFunc(res, err);
    });
}
export function Send(uri, request, resType, net, ...opts) {
    return __awaiter(this, void 0, void 0, function* () {
        if (net.is401()) {
            return [new resType(), new CodeError("has 401", Code.TokenExpire)];
        }
        let token = yield net.getToken();
        // token为空，需要处理401，并且不执行真正的网络请求。
        if (token === Token.empty) {
            yield net.process401();
            return [new resType(), new CodeError("token is empty", Code.TokenExpire)];
        }
        let [ret, err] = yield SendNoToken(uri, request, resType, net, ...opts);
        if (!err || err.code !== Code.TokenExpire) {
            // 取消401
            net.clear401();
            return [ret, err];
        }
        // 返回401时，如果请求的token与现在存储token不是一样的，说明有登录接口修改过，最新
        // 存储的token是否过期，无法判断，所以不能执行真正的401操作
        // 此处不能通过net层is401来判断，还没有开始处理401或者其他接口处理完401的情况下，is401都是false
        // 但是后一种情况却不能执行401；对token而言后一种情况token一定不一样，却可以分辨出两种情况
        if (token !== (yield net.getToken())) {
            return [new resType(), new CodeError("token is too old", Code.TokenExpire)];
        }
        yield net.process401();
        return [new resType(), new CodeError("token is expire", Code.TokenExpire)];
    });
}
